/* This file is part of the iox-ili project.
 * For more information, please see <http://www.eisenhutinformatik.ch/iox-ili/>.
 *
 * Copyright (c) 2006 Eisenhut Informatik AG
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included 
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */
package ch.interlis.ilirepository;

import java.io.File;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import ch.ehi.basics.logging.EhiLogger;
import ch.ehi.basics.tools.TopoSort;
import ch.interlis.ili2c.Ili2cException;
import ch.interlis.ili2c.config.Configuration;
import ch.interlis.ili2c.modelscan.IliFile;
import ch.interlis.ili2c.modelscan.IliModel;
import ch.interlis.ili2c.ModelScan;
import ch.interlis.ilirepository.impl.RepositoryAccess;
import ch.interlis.ilirepository.impl.RepositoryAccessException;
import ch.interlis.ilirepository.impl.RepositoryCrawler;
import ch.interlis.iom_j.xtf.XtfReader;
import ch.interlis.iox.IoxException;

import ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata_SchemaLanguage;
import ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata;

/** Enables cached access to a web of model repositories.
 * Usage:
 * <pre><code>
 * import ch.interlis.ili2c.Ili2c;
 * import ch.interlis.ili2c.config.Configuration;
 * import ch.interlis.ili2c.metamodel.TransferDescription;
 * // create repository manager
 * IliManager manager=new IliManager();
 * // set list of repositories to search
 * String repositories[]=new String[]{"http://models.interlis.ch/"};
 * manager.setRepositories(repositories);
 * // get list of ili-files for a given model
 * ArrayList modelNames=new ArrayList();
 * modelNames.add("DM01");
 * Configuration config=manager.getConfig(modelNames,1.0); // 1.0=ili 1 modell
 * // compile models
 * TransferDescritpion td=Ili2c.runCompiler(config);
 * </code></pre>
 * @author ceis
 */
public class IliManager {
	/** name of ilisite.xml file.
	 */
	public static final String ILISITE_XML = "ilisite.xml";
	/** name of ilimodels.xml file.
	 */
	public static final String ILIMODELS_XML = "ilimodels.xml";
	private long iliMaxTTL=604800000L; // max time to live in ms for a file in the cache
	private RepositoryAccess rep=new RepositoryAccess();
	private RepositoryCrawler crawler=new RepositoryCrawler();
	/** sets the list of repositories to start search for models.
	 */
	public void setRepositories(String uri[]){
		crawler.setup(uri,rep);
	}
	public void setIliFiles(String uri,IliFiles iliFiles){
		if(uri!=null && iliFiles!=null){
			rep.setIliFiles(uri, iliFiles);
		}
	}
	/** sets the folder used to cache files from remote repositories.
	 */
	public void setCache(File localdir){
		rep.setCache(localdir);
	}
	/** creates a compiler configuration, given a set of local ili-filepaths.
	 * @param requiredIliFileNames list<String iliFilename> list of local filenames
	 */
	public Configuration getConfigWithFiles(ArrayList requiredIliFileNames)
	throws Ili2cException
	{
		if(requiredIliFileNames.isEmpty()){
			throw new Ili2cException("no ili files given");
		}
		HashSet ilifiles=new HashSet();
		HashSet toVisitFiles=new HashSet(); // set<IliFile>
		HashSet requiredFiles=new HashSet(); // set<File>
		
		// scan given files and report about duplicate models
		Iterator reqFileIt=requiredIliFileNames.iterator();
		HashSet availablemodels=new HashSet();
		double version=0.0;
		while(reqFileIt.hasNext()){
			String fname=(String)reqFileIt.next();
			File file=new File(fname);
			if(!file.exists()){
				throw new Ili2cException(fname + ": " + "There is no such file.");
			}
			IliFile iliFile=ModelScan.scanIliFile(file);
			if(iliFile==null){
				throw new Ili2cException(fname + ": " + "Failed to scan ili-file.");
			}
			if(iliFile!=null){
				boolean skipThisFile=false;
				for(Iterator modeli=iliFile.iteratorModel();modeli.hasNext();){
					IliModel model=(IliModel)modeli.next();
					if(version==0.0){
						version=model.getIliVersion();
					}else{
						if(version!=model.getIliVersion()){
							skipThisFile=true;
							EhiLogger.logAdaption("different ili version; file ignored "+iliFile.getFilename());
							break;
						}
					}
					if(availablemodels.contains(model.getName())){
						skipThisFile=true;
						EhiLogger.logAdaption("duplicate model; file ignored "+iliFile.getFilename());
						break;
					}
				}
				if(!skipThisFile){
					requiredFiles.add(file);
					ilifiles.add(iliFile);
					toVisitFiles.add(iliFile);
					for(Iterator modeli=iliFile.iteratorModel();modeli.hasNext();){
						IliModel model=(IliModel)modeli.next();
						availablemodels.add(model.getName());
					}
				}
			}
		}

		return createConfig(toVisitFiles,version);
	}
	/** creates a compiler configuration, given a set of model names and ili-version.
	 * @param requiredModels list<String modelname> list of model names
	 * @param iliVersion ili version (e.g. 1.0) or 0.0 for auto-detection.
	 * @return compiler configuration
	 * @throws Ili2cException
	 */
	public Configuration getConfig(ArrayList requiredModels,double iliVersion)
	throws Ili2cException
	{
		HashSet toVisitFiles=new HashSet();
		
		// auto determine version?
		if(iliVersion==0.0){
			// get version of first model
			String firstModel=(String)requiredModels.get(0);
			IliModel model=null;
			IliFile ilifile=crawler.getIliFileMetadataDeep(firstModel,0.0,true);
			if(ilifile!=null){
				Iterator modeli=ilifile.iteratorModel();
				while(modeli.hasNext()){
					IliModel model1=(IliModel)modeli.next();
					if(model1.getName().equals(firstModel)){				
						iliVersion=model1.getIliVersion();
						model=model1;
						toVisitFiles.add(ilifile);
						break;
					}
				}
			}
			if(model==null){				
				// could not find model
				throw new Ili2cException(firstModel+": model not found");
			}
		}
		
		
		// build list of files with required models
		Iterator it=requiredModels.iterator();
		HashSet missingModels=new HashSet();
		StringBuffer errmsg=new StringBuffer();
		String sep="";
		while(it.hasNext()){
			String model=(String)it.next();
			if(model==null){
				continue;
			}
			IliFile file=getFromSet(toVisitFiles,model,iliVersion);
			if(file==null){
				file=crawler.getIliFileMetadataDeep(model,iliVersion,true);
				if(file==null){
					if(!missingModels.contains(model)){
						missingModels.add(model);
						errmsg.append(sep);
						errmsg.append(model);
						sep=", ";
					}
				}else{
					toVisitFiles.add(file);
				}
			}
		}	
		if(!missingModels.isEmpty()){
			errmsg.append(": model(s) not found");
			throw new Ili2cException(errmsg.toString());
		}
		
		// build config
		return createConfig(toVisitFiles,iliVersion);
	}
	/**
	 * 
	 * @param toVisitFiles set<IliFile iliFile>
	 * @return
	 */
	private Configuration createConfig(HashSet toVisitFiles,double iliVersion)
	throws Ili2cException
	{	
		if(toVisitFiles.isEmpty()){
			throw new IllegalStateException("toVisitFiles.isEmpty()");
		}
		HashSet visitedFiles=new HashSet();
		TopoSort reqFiles=new TopoSort();
		HashSet missingModels=new HashSet();
		StringBuffer errmsg=new StringBuffer();
		String sep="";
		while(!toVisitFiles.isEmpty()){
			IliFile ilifile=(IliFile)toVisitFiles.iterator().next();
			reqFiles.add(ilifile);

			Iterator modeli=ilifile.iteratorModel();
			while(modeli.hasNext()){
				IliModel model=(IliModel)modeli.next();
				Iterator depi=model.getDependencies().iterator();
				while(depi.hasNext()){
					String dep=(String)depi.next();
					//EhiLogger.debug("model "+model.getName()+", dep "+dep);
					IliFile supplier=getFromSet(visitedFiles,dep, iliVersion);
					if(supplier==null){
						supplier=getFromSet(toVisitFiles,dep, iliVersion);
					}
					if(supplier==null){
						supplier=crawler.getIliFileMetadataDeep(dep,iliVersion,true);
					}
					if(supplier==null){
						if(!missingModels.contains(dep)){
							errmsg.append(sep);
							errmsg.append(dep);
							sep=", ";
							missingModels.add(dep);
						}
					}else{
						if(!visitedFiles.contains(supplier)){
							// add file with supplier model
							toVisitFiles.add(supplier);
						}
						if(supplier!=ilifile){
							reqFiles.addcond(supplier,ilifile);
						}
					}
				}
			}
			toVisitFiles.remove(ilifile);
			visitedFiles.add(ilifile);
		}
		if(!missingModels.isEmpty()){
			errmsg.append(": model(s) not found");
			throw new Ili2cException(errmsg.toString());
		}
		if(!reqFiles.sort()){
			StringBuffer loopele=new StringBuffer();
			Iterator resi=reqFiles.getResult().iterator();
			sep="";
			while(resi.hasNext()){
			  IliFile res=(IliFile)resi.next();
			  loopele.append(sep);
			  loopele.append(res.getFilename().getName());
			  sep="->";
			}
			throw new Ili2cException("loop in ili-files: "+loopele.toString());
		}
		Iterator resi=reqFiles.getResult().iterator();
		Configuration config=new Configuration();
		while(resi.hasNext()){
		  IliFile res=(IliFile)resi.next();
		  // fetch file from repository to cache
		  File iliFile=null;
			try {
				iliFile = rep.getLocalFileLocation(res.getRepositoryUri(),res.getPath(),iliMaxTTL,res.getMd5());
				if(iliFile==null){
					if(res.getRepositoryUri()==null){
						throw new Ili2cException("failed to get ili file "+res.getPath());
					}
					throw new Ili2cException("failed to get ili file "+res.getRepositoryUri()+"/"+res.getPath());
				}
			} catch (RepositoryAccessException e) {
				if(res.getRepositoryUri()==null){
					throw new Ili2cException("failed to get ili file "+res.getPath(),e);
				}
				throw new Ili2cException("failed to get ili file "+res.getRepositoryUri()+"/"+res.getPath(),e);
			}
		  config.addFileEntry(new ch.interlis.ili2c.config.FileEntry(
				  iliFile.getPath(),ch.interlis.ili2c.config.FileEntryKind.ILIMODELFILE));
		}
		return config;
	}
	private IliFile getFromSet(HashSet ilifiles,String modelName,double iliVersion)
	{
		Iterator filei=ilifiles.iterator();
		while(filei.hasNext()){
			IliFile ilifile=(IliFile)filei.next();
			Iterator modeli=ilifile.iteratorModel();
			while(modeli.hasNext()){
				IliModel model=(IliModel)modeli.next();
				if(model.getName().equals(modelName) && model.getIliVersion()==iliVersion){				
					return ilifile;
				}
			}
		}
		return null;
		
	}
	static public void main(String args[])
	{
		EhiLogger.getInstance().setTraceFilter(false);
		IliManager m=new IliManager();
		m.setRepositories(new String[]{"C:/tmp/test23","http://models.interlis.ch"});
		ArrayList requiredModels=new ArrayList();
		ArrayList requiredFiles=new ArrayList();
		requiredModels.add("CatalogueObjects_V1");
		//requiredModels.add("Test23_erweitert");
		try {
			Configuration config=m.getConfig(requiredModels,0.0);
			if(config!=null){
				ch.interlis.ili2c.Ili2c.logIliFiles(config);
			}
		} catch (Ili2cException e) {
			EhiLogger.logError(e);
		}
		
	}
}

